import asyncio
import uuid
from typing import Any, Dict, Mapping, Optional, Union, cast

import aiohttp
from python_graphql_client import GraphqlClient
from starlette.datastructures import Headers

from chainlit.client.base import MessageDict, UserDict
from chainlit.config import config
from chainlit.logger import logger

from .base import BaseAuthClient, BaseDBClient, PageInfo, PaginatedResponse, UserDict


class GraphQLClient:
    def __init__(self, project_id: str, access_token: Optional[str]):
        self.project_id = project_id

        self.headers = {"content-type": "application/json"}
        if access_token:
            self.headers["Authorization"] = access_token

        graphql_endpoint = f"{config.chainlit_server}/api/graphql"
        self.graphql_client = GraphqlClient(
            endpoint=graphql_endpoint, headers=self.headers
        )

    async def query(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        :param query: The GraphQL query string.
        :param variables: A dictionary of variables for the query.
        :return: The response data as a dictionary.
        """
        return await self.graphql_client.execute_async(query=query, variables=variables)

    def check_for_errors(self, response: Dict[str, Any], raise_error: bool = False):
        if "errors" in response:
            if raise_error:
                raise Exception(response["errors"][0])
            logger.error(response["errors"][0])
            return True
        return False

    async def mutation(
        self, mutation: str, variables: Mapping[str, Any] = {}
    ) -> Dict[str, Any]:
        """
        Execute a GraphQL mutation.

        :param mutation: The GraphQL mutation string.
        :param variables: A dictionary of variables for the mutation.
        :return: The response data as a dictionary.
        """
        return await self.graphql_client.execute_async(
            query=mutation, variables=variables
        )


class CloudAuthClient(BaseAuthClient, GraphQLClient):
    def __init__(
        self,
        project_id: str,
        handshake_headers: Optional[Dict[str, str]] = None,
        request_headers: Optional[Headers] = None,
    ):
        access_token = None

        if handshake_headers:
            access_token = handshake_headers.get("HTTP_AUTHORIZATION")
        elif request_headers:
            access_token = request_headers.get("Authorization")

        if access_token is None:
            raise ConnectionRefusedError("No access token provided")

        GraphQLClient.__init__(self, project_id, access_token)

    async def get_user_infos(
        self,
    ) -> UserDict:
        data = {"projectId": self.project_id}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}/api/me",
                json=data,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    raise ValueError(f"Failed to get user infos. {r.status}: {reason}")

                json = await r.json()
                self.user_infos = json
                # replace with unpacking when Mypy 1.5.0 is out:
                # return UserDict(**self.user_infos)
                return cast(UserDict, self.user_infos)

    async def is_project_member(self) -> bool:
        try:
            user = await self.get_user_infos()
            return user["role"] != "ANONYMOUS"
        except ValueError as e:
            logger.error(e)
            return False


class CloudDBClient(BaseDBClient, GraphQLClient):
    conversation_id: Optional[str] = None
    lock: asyncio.Lock

    def __init__(
        self,
        project_id: str,
        handshake_headers: Optional[Dict[str, str]] = None,
        request_headers: Optional[Headers] = None,
    ):
        self.lock = asyncio.Lock()

        access_token = None

        if handshake_headers:
            access_token = handshake_headers.get("HTTP_AUTHORIZATION")
        elif request_headers:
            access_token = request_headers.get("Authorization")

        if access_token is None:
            raise ConnectionRefusedError("No access token provided")

        super().__init__(project_id, access_token)

    async def create_user(self, variables: UserDict) -> bool:
        raise NotImplementedError

    async def get_project_members(self):
        query = """query ($projectId: String!) {
                    projectMembers(projectId: $projectId) {
                    edges {
                        cursor
                        node {
                        role
                        user {
                            email
                            name
                        }
                        }
                    }
                    }
                }"""
        variables = {"projectId": self.project_id}
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        members = []

        for edge in res["data"]["projectMembers"]["edges"]:
            node = edge["node"]
            role = node["role"]
            name = node["user"]["name"]
            email = node["user"]["email"]
            members.append({"role": role, "name": name, "email": email})

        return members

    async def create_conversation(self) -> Optional[str]:
        # If we run multiple send concurrently, we need to make sure we don't create multiple conversations.
        async with self.lock:
            if self.conversation_id:
                return self.conversation_id

            mutation = """
            mutation ($projectId: String!, $sessionId: String) {
                createConversation(projectId: $projectId, sessionId: $sessionId) {
                    id
                }
            }
            """
            variables = {"projectId": self.project_id}
            res = await self.mutation(mutation, variables)

            if self.check_for_errors(res):
                logger.warning("Could not create conversation.")
                return None

            return res["data"]["createConversation"]["id"]

    async def get_conversation_id(self):
        self.conversation_id = await self.create_conversation()

        return self.conversation_id

    async def delete_conversation(self, conversation_id: str):
        mutation = """
        mutation ($id: ID!) {
            deleteConversation(id: $id) {
                id
            }
        }
        """
        variables = {"id": conversation_id}
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_conversation(self, conversation_id: str):
        query = """
        query ($id: ID!) {
            conversation(id: $id) {
                id
                createdAt
                messages {
                    id
                    isError
                    parentId
                    indent
                    author
                    content
                    waitForAnswer
                    humanFeedback
                    language
                    prompt
                    authorIsUser
                    createdAt
                }
                elements {
                    id
                    conversationId
                    type
                    name
                    url
                    display
                    language
                    size
                    forIds
                }
            }
        }
        """
        variables = {
            "id": conversation_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["conversation"]

    async def get_conversations(self, pagination, filter):
        query = """query (
        $first: Int
        $projectId: String!
        $cursor: String
        $withFeedback: Int
        $authorEmail: String
        $search: String
    ) {
        conversations(
        first: $first
        cursor: $cursor
        projectId: $projectId
        withFeedback: $withFeedback
        authorEmail: $authorEmail
        search: $search
        ) {
        pageInfo {
            endCursor
            hasNextPage
        }
        edges {
            cursor
            node {
            id
            createdAt
            elementCount
            messageCount
            author {
                name
                email
            }
            messages {
                content
            }
            }
        }
        }
    }"""

        variables = {
            "projectId": self.project_id,
            "first": pagination.first,
            "cursor": pagination.cursor,
            "withFeedback": filter.feedback,
            "authorEmail": filter.authorEmail,
            "search": filter.search,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        conversations = []

        for edge in res["data"]["conversations"]["edges"]:
            node = edge["node"]
            conversations.append(node)

        page_info = res["data"]["conversations"]["pageInfo"]

        return PaginatedResponse(
            pageInfo=PageInfo(
                hasNextPage=page_info["hasNextPage"],
                endCursor=page_info["endCursor"],
            ),
            data=conversations,
        )

    async def set_human_feedback(self, message_id, feedback):
        mutation = """mutation ($messageId: ID!, $humanFeedback: Int!) {
                        setHumanFeedback(messageId: $messageId, humanFeedback: $humanFeedback) {
                            id
                            humanFeedback
                    }
                }"""
        variables = {"messageId": message_id, "humanFeedback": feedback}
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_message(self):
        raise NotImplementedError

    async def create_message(self, variables: MessageDict) -> Optional[str]:
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the message.")
            return None

        variables["conversationId"] = c_id

        mutation = """
        mutation ($id: ID!, $conversationId: ID!, $author: String!, $content: String!, $language: String, $prompt: Json, $isError: Boolean, $parentId: String, $indent: Int, $authorIsUser: Boolean, $waitForAnswer: Boolean, $createdAt: StringOrFloat) {
            createMessage(id: $id, conversationId: $conversationId, author: $author, content: $content, language: $language, prompt: $prompt, isError: $isError, parentId: $parentId, indent: $indent, authorIsUser: $authorIsUser, waitForAnswer: $waitForAnswer, createdAt: $createdAt) {
                id
            }
        }
        """
        res = await self.mutation(mutation, variables)
        if self.check_for_errors(res):
            logger.warning("Could not create message.")
            return None

        return res["data"]["createMessage"]["id"]

    async def update_message(self, message_id: str, variables: MessageDict) -> bool:
        mutation = """
        mutation ($messageId: ID!, $author: String!, $content: String!, $parentId: String, $language: String, $prompt: Json) {
            updateMessage(messageId: $messageId, author: $author, content: $content, parentId: $parentId, language: $language,  prompt: $prompt) {
                id
            }
        }
        """
        res = await self.mutation(mutation, dict(messageId=message_id, **variables))

        if self.check_for_errors(res):
            logger.warning("Could not update message.")
            return False

        return True

    async def delete_message(self, message_id: str) -> bool:
        mutation = """
        mutation ($messageId: ID!) {
            deleteMessage(messageId: $messageId) {
                id
            }
        }
        """
        res = await self.mutation(mutation, {"messageId": message_id})

        if self.check_for_errors(res):
            logger.warning("Could not delete message.")
            return False

        return True

    async def get_element(self, conversation_id, element_id):
        query = """query (
        $conversationId: ID!
        $id: ID!
    ) {
        element(
        conversationId: $conversationId,
        id: $id
        ) {
        id
        conversationId
        type
        name
        url
        display
        language
        size
        forIds
        }
    }"""

        variables = {
            "conversationId": conversation_id,
            "id": element_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["element"]

    async def create_element(self, variables):
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the element.")
            return None

        mutation = """
        mutation ($conversationId: ID!, $type: String!, $name: String!, $display: String!, $forIds: [String!]!, $url: String, $objectKey: String, $size: String, $language: String) {
            createElement(conversationId: $conversationId, type: $type, url: $url, objectKey: $objectKey, name: $name, display: $display, size: $size, language: $language, forIds: $forIds) {
                id,
                type,
                url,
                objectKey,
                name,
                display,
                size,
                language,
                forIds
            }
        }
        """
        variables["conversationId"] = c_id
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not create element.")
            return None

        return res["data"]["createElement"]

    async def update_element(self, variables):
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the element.")
            return None

        mutation = """
        mutation ($conversationId: ID!, $id: ID!, $forIds: [String!]!) {
            updateElement(conversationId: $conversationId, id: $id, forIds: $forIds) {
                id,
            }
        }
        """

        variables = variables.copy()
        variables["conversationId"] = c_id
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not update element.")
            return None

        return res["data"]["updateElement"]

    async def upload_element(self, content: Union[bytes, str], mime: str) -> Dict:
        id = str(uuid.uuid4())
        body = {"projectId": self.project_id, "fileName": id, "contentType": mime}

        path = f"/api/upload/file"

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}{path}",
                json=body,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return {"object_key": None, "url": None}
                json_res = await r.json()

        upload_details = json_res["post"]
        object_key = upload_details["fields"]["key"]
        signed_url = json_res["signedUrl"]

        form_data = aiohttp.FormData()

        # Add fields to the form_data
        for field_name, field_value in upload_details["fields"].items():
            form_data.add_field(field_name, field_value)

        # Add file to the form_data
        form_data.add_field("file", content, content_type="multipart/form-data")
        async with aiohttp.ClientSession() as session:
            async with session.post(
                upload_details["url"],
                data=form_data,
            ) as upload_response:
                if not upload_response.ok:
                    reason = await upload_response.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return {"object_key": None, "url": None}

                url = f'{upload_details["url"]}/{object_key}'
                return {"object_key": object_key, "url": signed_url}



from typing import Dict, Optional

from fastapi import HTTPException, Request
from starlette.datastructures import Headers

from chainlit.client.base import BaseAuthClient, BaseDBClient, UserDict
from chainlit.client.cloud import CloudAuthClient, CloudDBClient
from chainlit.client.local import LocalAuthClient, LocalDBClient
from chainlit.config import config


async def get_auth_client(
    handshake_headers: Optional[Dict[str, str]] = None,
    request_headers: Optional[Headers] = None,
) -> BaseAuthClient:
    auth_client: Optional[BaseAuthClient] = None
    if config.code.auth_client_factory:
        auth_client = await config.code.auth_client_factory(
            handshake_headers, request_headers
        )
    elif not config.project.public and config.project.id:
        # Create the auth cloud client
        auth_client = CloudAuthClient(
            project_id=config.project.id,
            handshake_headers=handshake_headers,
            request_headers=request_headers,
        )

    if auth_client:
        # Check if the user is a member of the project
        is_project_member = await auth_client.is_project_member()
        if not is_project_member:
            raise HTTPException(
                status_code=401, detail="User is not a member of the project"
            )

        return auth_client

    # Default to local auth client
    return LocalAuthClient()


async def get_db_client(
    handshake_headers: Optional[Dict[str, str]] = None,
    request_headers: Optional[Headers] = None,
    user_infos: Optional[UserDict] = None,
) -> BaseDBClient:
    # Create the database client
    if config.project.database == "cloud":
        if not config.project.id:
            raise ValueError("Project id is required for database mode 'cloud'")

        return CloudDBClient(
            project_id=config.project.id,
            handshake_headers=handshake_headers,
            request_headers=request_headers,
        )
    elif config.project.database == "local":
        return LocalDBClient(user_infos=user_infos)
    elif config.project.database == "custom":
        if not config.code.db_client_factory:
            raise ValueError("Db client factory not provided")

        custom_db_client = await config.code.db_client_factory(
            handshake_headers, request_headers, user_infos
        )
        return custom_db_client

    raise ValueError("Unknown database type")


async def get_auth_client_from_request(
    request: Request,
) -> BaseAuthClient:
    # Get the auth client
    auth_client = await get_auth_client(None, request.headers)

    return auth_client


async def get_db_client_from_request(
    request: Request,
) -> BaseDBClient:
    # Get the auth client
    auth_client = await get_auth_client(None, request.headers)

    # Get the db client
    db_client = await get_db_client(None, request.headers, auth_client.user_infos)

    return db_client



import uuid
from typing import Any, Dict, Union, cast

from socketio.exceptions import TimeoutError

from chainlit.client.base import BaseDBClient, MessageDict
from chainlit.message import Message
from chainlit.session import Session
from chainlit.types import AskSpec, FileSpec


class ChainlitEmitter:
    """
    Chainlit Emitter class. The Emitter is not directly exposed to the developer.
    Instead, the developer interacts with the Emitter through the methods and classes exposed in the __init__ file.
    """

    session: Session

    def __init__(self, session: Session) -> None:
        """Initialize with the user session."""
        self.session = session

    def _get_session_property(self, property_name: str, raise_error=True):
        """Helper method to get a property from the session."""
        if not hasattr(self, "session") or not hasattr(self.session, property_name):
            if raise_error:
                raise ValueError(f"Session does not have property '{property_name}'")
            else:
                return None
        return getattr(self.session, property_name)

    @property
    def emit(self):
        """Get the 'emit' property from the session."""
        return self._get_session_property("emit")

    @property
    def ask_user(self):
        """Get the 'ask_user' property from the session."""
        return self._get_session_property("ask_user")

    @property
    def db_client(self) -> Union[BaseDBClient, None]:
        """Get the 'client' property from the session."""
        return self._get_session_property("db_client", raise_error=False)

    def send_message(self, msg_dict: Dict):
        """Send a message to the UI."""
        return self.emit("new_message", msg_dict)

    def update_message(self, msg_dict: Dict):
        """Update a message in the UI."""

        return self.emit("update_message", msg_dict)

    def delete_message(self, msg_dict):
        """Delete a message in the UI."""

        return self.emit("delete_message", msg_dict)

    def send_ask_timeout(self):
        """Send a prompt timeout message to the UI."""

        return self.emit("ask_timeout", {})

    def clear_ask(self):
        """Clear the prompt from the UI."""

        return self.emit("clear_ask", {})

    def enable_file_upload(self, spec: FileSpec):
        """Enable uploading file in the UI."""

        return self.emit("enable_file_upload", spec.to_dict())

    async def process_user_message(self, message_dict: MessageDict) -> Message:
        # Temporary UUID generated by the frontend should use v4
        assert uuid.UUID(message_dict["id"]).version == 4

        if self.db_client:
            # We have to update the UI with the actual DB ID
            ui_message_update = cast(Dict, message_dict.copy())
            persisted_id = await self.db_client.create_message(message_dict)
            if persisted_id:
                message_dict["id"] = persisted_id
                ui_message_update["newId"] = message_dict["id"]
                await self.update_message(ui_message_update)

        message = Message.from_dict(message_dict)
        self.session.root_message = message
        return message

    async def send_ask_user(
        self, msg_dict: Dict, spec: AskSpec, raise_on_timeout=False
    ):
        """Send a prompt to the UI and wait for a response."""

        try:
            # Send the prompt to the UI
            res = await self.ask_user(
                {"msg": msg_dict, "spec": spec.to_dict()}, spec.timeout
            )

            # End the task temporarily so that the User can answer the prompt
            await self.task_end()

            if self.db_client and res:
                # If cloud is enabled, store the response in the database/S3
                if spec.type == "text":
                    await self.process_user_message(res)
                elif spec.type == "file":
                    # TODO: upload file to S3
                    pass

            await self.clear_ask()
            return res
        except TimeoutError as e:
            await self.send_ask_timeout()

            if raise_on_timeout:
                raise e
        finally:
            await self.task_start()

    def update_token_count(self, count: int):
        """Update the token count for the UI."""

        return self.emit("token_usage", count)

    def task_start(self):
        """
        Send a task start signal to the UI.
        """
        return self.emit("task_start", {})

    def task_end(self):
        """Send a task end signal to the UI."""
        return self.emit("task_end", {})

    def stream_start(self, msg_dict: Dict):
        """Send a stream start signal to the UI."""
        return self.emit(
            "stream_start",
            msg_dict,
        )

    def send_token(self, id: str, token: str, is_sequence=False):
        """Send a message token to the UI."""
        return self.emit(
            "stream_token", {"id": id, "token": token, "isSequence": is_sequence}
        )

    def set_chat_settings(self, settings: Dict[str, Any]):
        self.session.chat_settings = settings



from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

from dataclasses_json import DataClassJsonMixin
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from chainlit.prompt import Prompt

InputWidgetType = Literal[
    "switch", "slider", "select", "textinput", "tags", "numberinput"
]
ElementType = Literal[
    "image", "avatar", "text", "pdf", "tasklist", "audio", "video", "file"
]
ElementDisplay = Literal["inline", "side", "page"]
ElementSize = Literal["small", "medium", "large"]


@dataclass
class FileSpec(DataClassJsonMixin):
    accept: Union[List[str], Dict[str, List[str]]]
    max_files: int
    max_size_mb: int


@dataclass
class AskSpec(DataClassJsonMixin):
    """Specification for asking the user."""

    timeout: int
    type: Literal["text", "file"]


@dataclass
class AskFileSpec(FileSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user a file."""


class AskResponse(TypedDict):
    content: str
    author: str


@dataclass
class AskFileResponse:
    name: str
    path: str
    size: int
    type: str
    content: bytes


class CompletionRequest(BaseModel):
    prompt: Prompt
    userEnv: Dict[str, str]


class UpdateFeedbackRequest(BaseModel):
    messageId: str
    feedback: Literal[-1, 0, 1]


class DeleteConversationRequest(BaseModel):
    conversationId: str


class Pagination(BaseModel):
    first: int
    cursor: Any


class ConversationFilter(BaseModel):
    feedback: Optional[Literal[-1, 0, 1]]
    authorEmail: Optional[str]
    search: Optional[str]


class GetConversationsRequest(BaseModel):
    pagination: Pagination
    filter: ConversationFilter



import json
import mimetypes

mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("text/css", ".css")

import asyncio
import os
import webbrowser
from contextlib import asynccontextmanager
from pathlib import Path

from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import FileResponse, HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi_socketio import SocketManager
from starlette.middleware.cors import CORSMiddleware
from watchfiles import awatch

from chainlit.client.utils import (
    get_auth_client_from_request,
    get_db_client_from_request,
)
from chainlit.config import DEFAULT_HOST, config, load_module, reload_config
from chainlit.logger import logger
from chainlit.markdown import get_markdown_str
from chainlit.playground.config import get_llm_providers
from chainlit.telemetry import trace_event
from chainlit.types import (
    CompletionRequest,
    DeleteConversationRequest,
    GetConversationsRequest,
    UpdateFeedbackRequest,
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    host = config.run.host
    port = config.run.port

    if host == DEFAULT_HOST:
        url = f"http://localhost:{port}"
    else:
        url = f"http://{host}:{port}"

    logger.info(f"Your app is available at {url}")

    if not config.run.headless:
        # Add a delay before opening the browser
        await asyncio.sleep(1)
        webbrowser.open(url)

    if config.project.database == "local":
        from prisma import Client, register  # type: ignore[attr-defined]

        client = Client()
        register(client)
        await client.connect()

    watch_task = None
    stop_event = asyncio.Event()

    if config.run.watch:

        async def watch_files_for_changes():
            extensions = [".py"]
            files = ["chainlit.md", "config.toml"]
            async for changes in awatch(config.root, stop_event=stop_event):
                for change_type, file_path in changes:
                    file_name = os.path.basename(file_path)
                    file_ext = os.path.splitext(file_name)[1]

                    if file_ext.lower() in extensions or file_name.lower() in files:
                        logger.info(
                            f"File {change_type.name}: {file_name}. Reloading app..."
                        )

                        try:
                            reload_config()
                        except Exception as e:
                            logger.error(f"Error reloading config: {e}")
                            break

                        # Reload the module if the module name is specified in the config
                        if config.run.module_name:
                            try:
                                load_module(config.run.module_name)
                            except Exception as e:
                                logger.error(f"Error reloading module: {e}")
                                break

                        await socket.emit("reload", {})

                        break

        watch_task = asyncio.create_task(watch_files_for_changes())

    try:
        yield
    finally:
        if config.project.database == "local":
            await client.disconnect()
        if watch_task:
            try:
                stop_event.set()
                watch_task.cancel()
                await watch_task
            except asyncio.exceptions.CancelledError:
                pass

        # Force exit the process to avoid potential AnyIO threads still running
        os._exit(0)


root_dir = os.path.dirname(os.path.abspath(__file__))
build_dir = os.path.join(root_dir, "frontend/dist")

app = FastAPI(lifespan=lifespan)

app.mount("/public", StaticFiles(directory="public", check_dir=False), name="public")
app.mount(
    "/assets",
    StaticFiles(
        packages=[("chainlit", os.path.join(build_dir, "assets"))],
        follow_symlink=config.project.follow_symlink,
    ),
    name="assets",
)


app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Define max HTTP data size to 100 MB
max_message_size = 100 * 1024 * 1024

socket = SocketManager(
    app,
    cors_allowed_origins=[],
    async_mode="asgi",
    max_http_buffer_size=max_message_size,
)


# -------------------------------------------------------------------------------
#                               HTTP HANDLERS
# -------------------------------------------------------------------------------


def get_html_template():
    PLACEHOLDER = "<!-- TAG INJECTION PLACEHOLDER -->"
    JS_PLACEHOLDER = "<!-- JS INJECTION PLACEHOLDER -->"

    default_url = "https://github.com/Chainlit/chainlit"
    url = config.ui.github or default_url

    tags = f"""<title>{config.ui.name}</title>
    <meta name="description" content="{config.ui.description}">
    <meta property="og:type" content="website">
    <meta property="og:title" content="{config.ui.name}">
    <meta property="og:description" content="{config.ui.description}">
    <meta property="og:image" content="https://chainlit-cloud.s3.eu-west-3.amazonaws.com/logo/chainlit_banner.png">
    <meta property="og:url" content="{url}">"""

    js = None
    if config.ui.theme:
        js = f"""<script>window.theme = {json.dumps(config.ui.theme.to_dict())}</script>"""

    index_html_file_path = os.path.join(build_dir, "index.html")

    with open(index_html_file_path, "r", encoding="utf-8") as f:
        content = f.read()
        content = content.replace(PLACEHOLDER, tags)
        if js:
            content = content.replace(JS_PLACEHOLDER, js)
        return content


@app.post("/completion")
async def completion(request: CompletionRequest):
    """Handle a completion request from the prompt playground."""

    providers = get_llm_providers()

    try:
        provider = [p for p in providers if p.id == request.prompt.provider][0]
    except IndexError:
        raise HTTPException(
            status_code=404,
            detail=f"LLM provider '{request.prompt.provider}' not found",
        )

    trace_event("pp_create_completion")
    response = await provider.create_completion(request)

    return response


@app.get("/project/llm-providers")
async def get_providers():
    """List the providers."""
    trace_event("pp_get_llm_providers")
    providers = get_llm_providers()
    providers = [p.to_dict() for p in providers]
    return JSONResponse(content={"providers": providers})


@app.get("/project/settings")
async def project_settings():
    """Return project settings. This is called by the UI before the establishing the websocket connection."""
    return JSONResponse(
        content={
            "chainlitServer": config.chainlit_server,
            "prod": bool(config.chainlit_prod_url),
            "ui": config.ui.to_dict(),
            "project": config.project.to_dict(),
            "markdown": get_markdown_str(config.root),
        }
    )


@app.put("/message/feedback")
async def update_feedback(request: Request, update: UpdateFeedbackRequest):
    """Update the human feedback for a particular message."""

    db_client = await get_db_client_from_request(request)
    await db_client.set_human_feedback(
        message_id=update.messageId, feedback=update.feedback
    )
    return JSONResponse(content={"success": True})


@app.get("/project/members")
async def get_project_members(request: Request):
    """Get all the members of a project."""

    db_client = await get_db_client_from_request(request)
    res = await db_client.get_project_members()
    return JSONResponse(content=res)


@app.get("/project/role")
async def get_member_role(request: Request):
    """Get the role of a member."""

    auth_client = await get_auth_client_from_request(request)
    role = auth_client.user_infos["role"] if auth_client.user_infos else "ANONYMOUS"
    return JSONResponse(content=role)


@app.post("/project/conversations")
async def get_project_conversations(request: Request, payload: GetConversationsRequest):
    """Get the conversations page by page."""

    db_client = await get_db_client_from_request(request)
    res = await db_client.get_conversations(payload.pagination, payload.filter)
    return JSONResponse(content=res.to_dict())


@app.get("/project/conversation/{conversation_id}")
async def get_conversation(request: Request, conversation_id: str):
    """Get a specific conversation."""

    db_client = await get_db_client_from_request(request)
    res = await db_client.get_conversation(conversation_id)
    return JSONResponse(content=res)


@app.get("/project/conversation/{conversation_id}/element/{element_id}")
async def get_conversation_element(
    request: Request, conversation_id: str, element_id: str
):
    """Get a specific conversation element."""

    db_client = await get_db_client_from_request(request)
    res = await db_client.get_element(conversation_id, element_id)
    return JSONResponse(content=res)


@app.delete("/project/conversation")
async def delete_conversation(request: Request, payload: DeleteConversationRequest):
    """Delete a conversation."""

    db_client = await get_db_client_from_request(request)
    await db_client.delete_conversation(conversation_id=payload.conversationId)
    return JSONResponse(content={"success": True})


@app.get("/files/{filename:path}")
async def serve_file(filename: str):
    base_path = Path(config.project.local_fs_path).resolve()
    file_path = (base_path / filename).resolve()

    # Check if the base path is a parent of the file path
    if base_path not in file_path.parents:
        raise HTTPException(status_code=400, detail="Invalid filename")

    if file_path.is_file():
        return FileResponse(file_path)
    else:
        raise HTTPException(status_code=404, detail="File not found")


@app.get("/favicon.svg")
async def get_favicon():
    favicon_path = os.path.join(build_dir, "favicon.svg")
    return FileResponse(favicon_path, media_type="image/svg+xml")


def register_wildcard_route_handler():
    @app.get("/{path:path}")
    async def serve(request: Request, path: str):
        html_template = get_html_template()
        """Serve the UI files."""
        response = HTMLResponse(content=html_template, status_code=200)

        key = "chainlit-initial-headers"

        chainlit_initial_headers = dict(request.headers)
        if "cookie" in chainlit_initial_headers:
            del chainlit_initial_headers["cookie"]

        response.set_cookie(
            key=key,
            value=json.dumps(chainlit_initial_headers),
            httponly=True,
        )

        return response


import chainlit.socket  # noqa
